/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	GLOBAL DEFINITIONS
 *
 *	These are some frequently used constants and structures.
 */

#define _BPW_		2		/* Bytes per word */
#define _BITS_		16		/* Bits per word */

#define START		"START"		/* Main function name (IEP) */

#define NEWLINE		"\r\n"		/* DOS: CR,LF  Unix: LF */

#define PATHL		81		/* Max. source file name length */
#define TOKLEN		128		/* Max. token length */
#define BUFLEN		512		/* Input buffer length, must be */
					/* at least TOKLEN*3 */
#define OBUFLEN		512		/* Output buffer length */
					/* Bigger is better */

#define NAMELEN		8		/* Max. length of a symbol name */
					/* Longer names will be truncated */

#define MEMTOP		32768U		/* Bytes reserved for nodes and */
					/* symbol table */
#define DFLNODES	1600		/* Default size of node pool */
#define MAXLBL		32766		/* Max. number of labels */
#define TBLLEN		512		/* Max. size of a TABLE */
#define MAXCASE		256		/* Max. number of cases in SWITCH */
#define MAXFORW		128		/* Max. number of FORWARD's */

#define DSMAX		32767		/* Size of static data segment */
#define STKLEN		30000		/* Default size of RT stack */

#define uint	unsigned

/* RELOCATION FLAGS */
#define DATA	1			/* Base = data segment */
#define CODE	2			/* Base = code segment */
#define NONE	0			/* Do not relocate */

/* SYMBOL TABLE ENTRY */
struct syment {
	char	*name;			/* Object name */
	uint	addr,			/* Address */
		size;			/* Size (for vectors) */
	char	level,			/* Nesting level */
		flags;			/* Flags, see below */
};

#define PUBLIC	0001		/* Symbol is public */
#define EXTERN	0002		/* Symbol is extern */
#define STATIC	0004		/* Object resides in static data segment */
#define FORWARD	0010		/* FORWARD pointer */
#define STRUCT	0020		/* Symbols starts a STRUCTURE */
#define RVALUE	0040		/* Object is a constant */
#define REFERD	0100		/* Object has been referenced */


/* PARSE TREE NODE */
struct node {
	int	token;			/* BCODE instruction */
	int	flags;			/* Flags, see below */
	union {
		int	val;		/* value or */
		struct syment	*sym;	/* symbol table pointer */
	} v;
	struct node	*left,		/* left child */
			*right;		/* right child */
	struct node	*nlink;		/* freelist link */
};

#define CONST	0001		/* Subtree is constant (rvalue) */
#define DIRECT	0002		/* Subtree is a directlty accessible symbol */
#define GLOBAL	0004		/* If DIRECT: symbol is global */
