/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	BCODE INSTRUCTIONS
 *
 *	This is the complete BCODE instruction set.
 *	Not all instructions are necessarily used by the compiler.
 */

#define GLUE	0000		/* glue */

#define G_ADD	0001		/* s1 s0 --> s1 + s0 */
#define G_SUB	0002		/* s1 s0 --> s1 - s0 */
#define G_MUL	0003		/* s1 s0 --> s1 * s0 */
#define G_DIV	0004		/* s1 s0 --> s1 / s0 */
#define G_MOD	0005		/* s1 s0 --> s1 % s0 */
#define G_NEG	0006		/* s0 --> -s0 */

#define G_AND	0010		/* s1 s0 --> s1 & s0 */
#define G_OR	0011		/* s1 s0 --> s1 | s0 */
#define G_XOR	0012		/* s1 s0 --> s1 ^ s0 */
#define G_SHL	0013		/* s1 s0 --> s1 << s0 */
#define G_SHR	0014		/* s1 s0 --> s1 >> s0 */
#define G_ROL	0015		/* s1 s0 --> s1 {{ s0 */
#define G_ROR	0016		/* s1 s0 --> s1 }} s0 */
#define G_COM	0017		/* s0 --> ~s0 */

#define G_EQ	0020		/* s1 s0 --> s1 == s0 */
#define G_NE	0021		/* s1 s0 --> s1 != s0 */
#define G_LT	0022		/* s1 s0 --> s1 < s0 */
#define G_GT	0023		/* s1 s0 --> s1 > s0 */
#define G_LE	0024		/* s1 s0 --> s1 <= s0 */
#define G_GE	0025		/* s1 s0 --> s1 >= s0 */
#define G_ULT	0026		/* s1 s0 --> s1 U< s0 */
#define G_UGT	0027		/* s1 s0 --> s1 U> s0 */
#define G_ULE	0030		/* s1 s0 --> s1 U<= s0 */
#define G_UGE	0031		/* s1 s0 --> s1 U>= s0 */
#define G_NOT	0032		/* s0 --> !s0 */

#define G_LN	0040		/* LN n */
#define G_LA	0041		/* LA a */
#define G_LC	0042		/* LC a */
#define G_LL	0043		/* LL r */
#define G_LG	0044		/* LG a */
#define G_LX	0045		/* LX n */
#define G_LXA	0046		/* LXA n */
#define G_LV	0047		/* s0 --> .s0 */
#define G_SL	0050		/* SL r */
#define G_SG	0051		/* SG a */
#define G_SX	0052		/* SX n */
#define G_SV	0053		/* s1 sa --> */
#define G_ADJ	0054		/* sa --> sa */
#define G_SAV	0055		/* n --> n n */

#define G_LBL	0060		/* LBL l */
#define G_JMP	0061		/* JMP l */
#define G_JFA	0062		/* JFA l */
#define G_JTR	0063		/* JTR l */
#define G_VJP	0064		/* sa --> */
#define G_SWI	0065		/* SWI n l n0 l0 ... */
#define G_HLT	0066		/* s0 --> */

#define G_CAL	0070		/* CAL l */
#define G_VCA	0071		/* sa --> */
#define G_FRM	0072		/* --> */
#define G_RTN	0073		/* sa --> */
#define G_PSH	0074		/* --> s0 */
#define G_RMA	0075		/* s? ... s0 --> */
#define G_VEC	0076		/* --> s? ... s0 */
#define G_UNV	0077		/* s? ... s0 --> */

#define G_INI	0100		/* INI a n ... */
#define G_REL	0101		/* ini . . R n ... */
#define G_RLA	0102		/* ini . . L n ... */
#define G_PUB	0103		/* PUB n name ... */
#define G_EXT	0104		/* EXT n name n ... */
#define G_END	0105		/* END n */
#define G_NAM	0106		/* NAM name */

#define G_IMG	0110		/* IMG n c ... */
#define G_DRO	0111		/* ? --> */

#define G_LIN	0120		/* LIN n (debugging) */

#define G_NUM	0175		/* tmp node: insert literal number */
#define G_SYM	0176		/* tmp node: symbol table pointer */
