/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	THE MAIN MODULE
 *
 *	Initialize, run, and shutdown the compiler.
 *
 *	Public functions: none
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#define extern
#include "data.h"
#undef extern
#include "error.h"
#include "token.h"
#include "glue.h"

#include "proto.h"


/*
 *	INIT(NNOD, NSYM)
 *	Purpose:	Initialize globals, allocate pool memory, and
 *			symbol table memory.
 *	Receives:	NNOD	- number of nodes to allocate
 *			NSYM	- number of symbol entries to allocate
 *	Returns:	-
 */
void init(nnod, nsym) {
	sc_init(NULL);
	em_init(NULL);
	ut_init(nnod);
	pa_init();
	sy_init(nsym);

	O_debug = 0;
	O_opt = 1;
	O_linfo = 0;
}


/*
 *	END()
 *	Purpose:	Shut down the compiler (flush buffers,
 *			dump symbol table [debug], free memory,
 *			close files).
 *	Receives:	-
 *	Returns:	-
 */
void end() {
	emit_sym(PUBLIC);
	emit_sym(EXTERN);
	t_emit(G_END);
	n_emit(Dtop);
	do_emit(NULL);
	if (O_debug) dumpsym();
	free(Nodes);
	free(Symbols);
}


/*
 *	OPTIONS()
 *	Purpose:	Scan an OPTIONS string, and set the requested
 *			compiler switches.
 *	Receives:	S	- An options string (the argument to
 *				  OPTIONS).
 *	Returns:	-
 */
void options(s)
char	*s;
{
	int	flag = 1;
	int	*optr = NULL;

	while (*s) {
		switch (*s) {
			case '+':
				flag = 1; break;
			case '-':
				flag = 0; break;
			case 'd': case 'D':
				O_debug = flag; optr = &O_debug; break;
			case 'l': case 'L':
				O_linfo = flag; optr = &O_linfo; break;
			case 'o': case 'O':
				O_opt = flag; optr = &O_opt; break;

			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				if (optr) *optr = (*s)-'0'; break;
		}
		s++;
	}
}


/*
 *	MAIN()
 *	Purpose:	startup.
 *	Receives:	command line arguments.
 *	Returns:	hopefully.
 */
main() {
	int	nnod, nsym;

	nnod = DFLNODES;
	nsym = (MEMTOP - nnod*sizeof(struct node)) / sizeof(struct syment);
	init(nnod, nsym);
	parse();
	end();
	return(Errcount? 1: O_debug);
}
