/******************************************************************************
 *
 *	The BSL Compiler Kit
 *	A SIMPLE i8086 CODE GENERATOR
 *	COPYRIGHT (C) 1995 NILS M. HOLM
 *
 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "../../boot/front/glue.h"


/******************************************************************************
 *	ASSEMBLY LANGUAGE OPCODES
 *****************************************************************************/
#define PUSHAX()	emitb(0x50)
#define POPAX()		emitb(0x58)
#define PUSHBP()	emitb(0x55)
#define POPBX()		emitb(0x5B)
#define POPCX()		emitb(0x59)
#define POPBP()		emitb(0x5D)
#define MOVBPSP()	emitw(0xEC8B)
#define MOVAXN()	emitb(0xB8)
#define MOVAXG()	emitb(0xA1)
#define MOVAXL(a)	if (near(a)) emitw(0x468B); else emitw(0x868B)
#define MOVAXDX()	emitw(0xC28B)
#define MOVCXN()	emitb(0xB9)
#define MOVCXG()	emitw(0x0E8B)
#define MOVCXL(a)	if (near(a)) emitw(0x4E8B); else emitw(0x8E8B)
#define MOVSIN()	emitb(0xBE)
#define ADDN()		emitb(0x05)
#define ADDL(a)		if (near(a)) emitw(0x4603); else emitw(0x8603)
#define ADDG()		emitw(0x0603)
#define ADDCX()		emitw(0xC103)
#define SUBN()		emitb(0x2d)
#define SUBL(a)		if (near(a)) emitw(0x462b); else emitw(0x862b)
#define SUBG()		emitw(0x062b)
#define SUBCX()		emitw(0xC12b)
#define MULCX()		emitw(0xE1F7)
#define XORDXDX()	emitw(0xD231)
#define DIVCX()		emitw(0xF1F7)
#define ANDN()		emitb(0x25)
#define ANDL(a)		if (near(a)) emitw(0x4623); else emitw(0x8623)
#define ANDG()		emitw(0x0623)
#define ANDCX()		emitw(0xC123)
#define ORN()		emitb(0x0D)
#define ORL(a)		if (near(a)) emitw(0x460B); else emitw(0x860B)
#define ORG()		emitw(0x060B)
#define ORCX()		emitw(0xC10B)
#define XORN()		emitb(0x35)
#define XORL(a)		if (near(a)) emitw(0x4633); else emitw(0x8633)
#define XORG()		emitw(0x0633)
#define XORCX()		emitw(0xC133)
#define SHLCL()		emitw(0xE0D3)
#define SHRCL()		emitw(0xE8D3)
#define ROLCL()		emitw(0xC0D3)
#define RORCL()		emitw(0xC8D3)
#define CMPN()		emitb(0x3D)
#define CMPL(a)		if (near(a)) emitw(0x463B); else emitw(0x863B)
#define CMPG()		emitw(0x063B)
#define CMPCXAX()	emitw(0xC83B)
#define JE()		emitb(0x74)
#define JNE()		emitb(0x75)
#define JL()		emitb(0x7C)
#define JG()		emitb(0x7F)
#define JLE()		emitb(0x7e)
#define JGE()		emitb(0x7d)
#define JA()		emitb(0x77)
#define JB()		emitb(0x72)
#define JAE()		emitb(0x73)
#define JBE()		emitb(0x76)
#define JMPS()		emitb(0xEB)
#define XORAX()		emitw(0xC033)
#define ORAX()		emitw(0xC00B)
#define XCHGAXCX()	emitb(0x91)
#define XCHG()		XCHGAXCX()
#define NEGAX()		emitw(0xD8F7)
#define NOTAX()		emitw(0xD0F7)
#define MOVBXAX()	emitw(0xD88B)
#define MOVAXP()	emitw(0x078B)
#define SBBAX()		emitw(0xC019)
#define INCAX()		emitb(0x40)
#define MOVLAX(a)	if (near(a)) emitw(0x4689); else emitw(0x8689)
#define MOVGAX()	emitb(0xA3)
#define MOVPAX()	emitw(0x0789)
#define SHLBX1()	emitw(0xE3D1)
#define ADDSPN(n)	if (near(n)) emitw(0xC483); else emitw(0xC481)
#define SUBSPN(n)	if (near(n)) emitw(0xEC83); else emitw(0xEC81)
#define RET()		emitb(0xC3)
#define CALLAX()	emitw(0xD0FF)
#define CALLG()		emitw(0x16FF)
#define CALLL(a)	if (near(a)) emitw(0x56FF); else emitw(0x96FF)
#define JMP()		emitb(0xE9)
#define SHLAX1()	emitw(0xE0D1)
#define MOVBXBP()	emitw(0xDD8B)
#define SHRBX1()	emitw(0xEBD1)
#define SUBBXAX()	emitw(0xC329)
#define MOVAXBX()	emitw(0xD889)
#define MOVAHN()	emitb(0xB4)
#define DOSINT()	emitw(0x21CD)
#define HALT()		do { MOVAHN(); emitb(0x4C); DOSINT(); } while(0)


/******************************************************************************
 *	DEFINITIONS ...
 *****************************************************************************/

#define _BPW_	2
#define SEGMENT	16
#define FORCE_SWITCH

#define uint	unsigned int
#define uchar	unsigned char

#define iswhite(c)	((c) == ' ' || (c) == '\n' || (c) == '\r')
#define near(a)		(-128 <= (int) (a) && (int) (a) <= 127)
#define emitb(b)	c_append(b, 1)
#define emitw(w)	c_append(w, 2)
#define emitv(a)	if (near(a)) emitb(a); else emitw(a)
#define demit(w)	d_append(w)
#define mkdisp(o, t)	((t) - (o)-2)
#define rel(a)		((a)+Dframe)
#define arel(a)		((a)+Dframe/2)

#define MAXREC	1280
#define BUFLEN	1024
#define STRLEN	128
#define STKLEN	64

#define ENDFILE	-1
#define NUMBER	-2
#define NAME	-3

#define BADLBL	((uint)-1)

#define TDATA	"cgd00000"
#define TCODE	"cgc00000"
#define MAP	"map"

struct label {
	int	id;
	uint	offset;
	struct label	*next;
};

struct extref {
	int	id;
	uint	offset;
	char	*name;
	struct extref	*next;
};

struct pubsym {
	char	*name;
	int	offset;
	struct pubsym	*next;
};

/******************************************************************************
 *	GLOBALS ...
 *****************************************************************************/

int	Acc_busy;
uchar	Cbuf[BUFLEN], Dbuf[BUFLEN];
uint	Cp, Dp;
uint	Ctop, Cframe, Dtop, Dframe;

char	Buffer[BUFLEN+STRLEN+1];
int	Btop, Bp, Lastrd;
int	Value;
char	Sname[STRLEN+1];
int	Istack[STKLEN], Astack[STKLEN], Sp;

int	Infile, Outfile;
int	Tmpdata, Tmpcode;
int	Line;

struct label	*Labels, *Cmarks, *Dmarks;

struct extref	*Erefs;

struct pubsym	*Pubtbl;

uint	Switch_entry;

char	*Module_name;


/******************************************************************************
 *	ERROR HANDLING
 *****************************************************************************/

void error(msg)
char	*msg;
{
	fprintf(stderr, "CG error in module %s (line %d): %s\n",
	Module_name? Module_name: "(UNNAMED)", Line, msg);
	exit(1);
}


void nomem() {
	error("out of memory");
}


void context(id) {
	char	buf[30];

	sprintf(buf, "bad context: %d", id);
	error(buf);
}


/******************************************************************************
 *	INITIALIZATION AND CLEANUP
 *****************************************************************************/

void init() {
	Infile = fileno(stdin);
	Outfile = fileno(stdout);
	if ((Tmpdata = creat(TDATA, S_IREAD|S_IWRITE)) < 0)
		error("cannot create tmp file");
	if ((Tmpcode = creat(TCODE, S_IREAD|S_IWRITE)) < 0)
		error("cannot create tmp file");
	setmode(Infile, O_BINARY);
	setmode(Outfile, O_BINARY);
	setmode(Tmpcode, O_BINARY);
	setmode(Tmpdata, O_BINARY);
	Lastrd = Btop = Bp = 0;
	Cframe = Ctop = 0;
	Dframe = Dtop = 0;
	Cp = Dp = 0;
	Sp = 0;
	Acc_busy = 0;
	Labels = Cmarks = Dmarks = NULL;
	Erefs = NULL;
	Switch_entry = BADLBL;
	Module_name = NULL;
	Line = 1;
}


void writemap() {
	struct pubsym	*pp;
	char	buf[81];
	int	mapfd;

	if ((mapfd = creat(MAP, S_IWRITE)) < 0) return;
	for (pp = Pubtbl; pp; pp = pp->next) {
		sprintf(buf, "%04X %04X %s\n",
		pp->offset, pp->offset/_BPW_, pp->name);
		write(mapfd, buf, strlen(buf));
	}
	close(mapfd);
}


void end() {
	if (Infile != fileno(stdin)) close(Infile);
	if (Outfile != fileno(stdout)) close(Outfile);
	if (!Erefs) writemap();
	close(Tmpcode);
	close(Tmpdata);
	unlink(TCODE);
	unlink(TDATA);
}


/******************************************************************************
 *	SCANNER / TOKENIZER
 *****************************************************************************/

readb() {
	if (Lastrd < 0) {
		return((Bp >= Btop)? -1: 0);
	}
	if (Bp != Btop && Btop > 0)
		memcpy(Buffer, &Buffer[Bp], Btop-Bp);
	Bp = Btop-Bp;
	if ((Btop = read(Infile, &Buffer[Bp], BUFLEN)) < 0)
		error("I/O Error reading file");
	if (Btop < BUFLEN) Lastrd = -1;
	Btop += Bp;
	Bp = 0;
	Buffer[Btop] = 0;
	return((!Btop)? -1: 0);
}


int readword(buf)
char	*buf;
{
	int	i;

	if (Bp >= Btop-STRLEN) {
		if (readb() < 0) return(ENDFILE);
	}
	while (iswhite(Buffer[Bp]) && Bp < Btop) {
		if (Buffer[Bp] == '\n') Line++;
		Bp++;
	}
	if (Bp >= Btop) return(ENDFILE);
	for (i=0; !iswhite(Buffer[Bp]) && Buffer[Bp]; )
		buf[i++] = Buffer[Bp++];
	buf[i] = 0;

	return(0);
}


next() {
	char	buf[STRLEN+1];

	if (readword(buf) == ENDFILE) return(ENDFILE);

	switch(*buf) {
		case 'A':
			if (!strcmp(buf, "ADD")) return(G_ADD);
			if (!strcmp(buf, "ADJ")) return(G_ADJ);
			if (!strcmp(buf, "AND")) return(G_AND);
			break;

		case 'C':
			if (!strcmp(buf, "CAL")) return(G_CAL);
			if (!strcmp(buf, "COM")) return(G_COM);
			break;

		case 'D':
			if (!strcmp(buf, "DIV")) return(G_DIV);
			if (!strcmp(buf, "DRO")) return(G_DRO);
			break;

		case 'E':
			if (!strcmp(buf, "END")) return(G_END);
			if (!strcmp(buf, "EQ")) return(G_EQ);
			if (!strcmp(buf, "EXT")) return(G_EXT);
			break;

		case 'F':
			if (!strcmp(buf, "FRM")) return(G_FRM);
			break;

		case 'G':
			if (!strcmp(buf, "GE")) return(G_GE);
			if (!strcmp(buf, "GT")) return(G_GT);
			break;

		case 'H':
			if (!strcmp(buf, "HLT")) return(G_HLT);
			break;

		case 'I':
			if (!strcmp(buf, "IMG")) return(G_IMG);
			if (!strcmp(buf, "INI")) return(G_INI);
			break;

		case 'J':
			if (!strcmp(buf, "JFA")) return(G_JFA);
			if (!strcmp(buf, "JMP")) return(G_JMP);
			if (!strcmp(buf, "JTR")) return(G_JTR);
			break;

		case 'L':
			if (!strcmp(buf, "LA")) return(G_LA);
			if (!strcmp(buf, "LBL")) return(G_LBL);
			if (!strcmp(buf, "LC")) return(G_LC);
			if (!strcmp(buf, "LE")) return(G_LE);
			if (!strcmp(buf, "LG")) return(G_LG);
			if (!strcmp(buf, "LIN")) return(G_LIN);
			if (!strcmp(buf, "LL")) return(G_LL);
			if (!strcmp(buf, "LN")) return(G_LN);
			if (!strcmp(buf, "LT")) return(G_LT);
			if (!strcmp(buf, "LV")) return(G_LV);
			if (!strcmp(buf, "LX")) return(G_LX);
			if (!strcmp(buf, "L")) return(G_RLA);
			break;

		case 'M':
			if (!strcmp(buf, "MOD")) return(G_MOD);
			if (!strcmp(buf, "MUL")) return(G_MUL);
			break;

		case 'N':
			if (!strcmp(buf, "NAM")) return(G_NAM);
			if (!strcmp(buf, "NE")) return(G_NE);
			if (!strcmp(buf, "NEG")) return(G_NEG);
			if (!strcmp(buf, "NOT")) return(G_NOT);
			if (!strcmp(buf, "NUM")) return(G_NUM);
			break;

		case 'O':
			if (!strcmp(buf, "OR")) return(G_OR);
			break;

		case 'P':
			if (!strcmp(buf, "PSH")) return(G_PSH);
			if (!strcmp(buf, "PUB")) return(G_PUB);
			break;

		case 'R':
			if (!strcmp(buf, "R")) return(G_REL);
			if (!strcmp(buf, "RMA")) return(G_RMA);
			if (!strcmp(buf, "ROL")) return(G_ROL);
			if (!strcmp(buf, "ROR")) return(G_ROR);
			if (!strcmp(buf, "RTN")) return(G_RTN);
			break;

		case 'S':
			if (!strcmp(buf, "SAV")) return(G_SAV);
			if (!strcmp(buf, "SG")) return(G_SG);
			if (!strcmp(buf, "SHL")) return(G_SHL);
			if (!strcmp(buf, "SHR")) return(G_SHR);
			if (!strcmp(buf, "SL")) return(G_SL);
			if (!strcmp(buf, "SUB")) return(G_SUB);
			if (!strcmp(buf, "SV")) return(G_SV);
			if (!strcmp(buf, "SWI")) return(G_SWI);
			if (!strcmp(buf, "SX")) return(G_SX);
			if (!strcmp(buf, "SYM")) return(G_SYM);
			break;

		case 'U':
			if (!strcmp(buf, "UGE")) return(G_UGE);
			if (!strcmp(buf, "UGT")) return(G_UGT);
			if (!strcmp(buf, "ULE")) return(G_ULE);
			if (!strcmp(buf, "ULT")) return(G_ULT);
			if (!strcmp(buf, "UNV")) return(G_UNV);
			break;

		case 'V':
			if (!strcmp(buf, "VCA")) return(G_VCA);
			if (!strcmp(buf, "VEC")) return(G_VEC);
			if (!strcmp(buf, "VJP")) return(G_VJP);
			break;

		case 'X':
			if (!strcmp(buf, "XOR")) return(G_XOR);
			break;
	}

	if (isdigit(*buf) || *buf == '-') {
		Value = atoi(buf);
		return(NUMBER);
	}

	strcpy(Sname, buf);
	return(NAME);
}


char *string() {
	char	buf[STRLEN+1];

	if (readword(buf) == ENDFILE) error("name expected (found EOF)");
	strcpy(Sname, buf);
	return(Sname);
}


num() {
	if (next() != NUMBER) error("numeric operand required");
	return(Value);
}


addr() {
	num();
	return(Value * _BPW_);
}


laddr() {
	num();
	return (Value<0? -(Value*_BPW_): -(Value*_BPW_));
}


/******************************************************************************
 *	CODE/DATA ARRAY MANAGEMENT
 *****************************************************************************/

void dwrite(fd, buf, len)
int	fd;
char	*buf;
int	len;
{
	if (write(fd, buf, len) != len)
		error("I/O error writing file");
}


void c_append(obj, len) {
	if (Cp + len >= BUFLEN) {
		dwrite(Tmpcode, Cbuf, Cp);
		Cp = 0;
	}
	Cbuf[Cp++] = obj & 0xff;
	if (len>1) Cbuf[Cp++] = obj>>8;
	Ctop += len;
}


uint c_extract(pos)
uint	pos;
{
	uint	buf, bbot = Ctop-Cp;

	if (pos >= bbot) return(Cbuf[pos-bbot] + (Cbuf[pos-bbot+1]<<8));

	lseek(Tmpcode, (long) pos, SEEK_SET);
	read(Tmpcode, &buf, _BPW_);
	lseek(Tmpcode, 0L, SEEK_END);
	return(buf);
}


void c_adjust(pos, new)
uint	pos;
int	new;
{
	uint	bbot = Ctop-Cp;

	if (pos >= bbot) {
		Cbuf[pos-bbot] = new&0xff;
		Cbuf[pos-bbot+1] = new>>8;
		return;
	}

	lseek(Tmpcode, (long) pos, SEEK_SET);
	dwrite(Tmpcode, &new, _BPW_);
	lseek(Tmpcode, 0L, SEEK_END);
}


void d_append(obj) {
	if (Dp + 2 >= BUFLEN) {
		dwrite(Tmpdata, Dbuf, Dp);
		Dp = 0;
	}
	Dbuf[Dp++] = obj & 0xff;
	Dbuf[Dp++] = obj>>8;
	Dtop += 2;
}


uint d_extract(pos)
uint	pos;
{
	uint	buf, bbot = Dtop-Dp;

	if (pos >= bbot) return(Dbuf[pos-bbot] + (Dbuf[pos-bbot+1]<<8));

	lseek(Tmpdata, (long) pos, SEEK_SET);
	read(Tmpdata, &buf, _BPW_);
	lseek(Tmpdata, 0L, SEEK_END);
	return(buf);
}


void d_adjust(pos, new)
uint	pos;
int	new;
{
	uint	bbot = Dtop-Dp;

	if (pos >= bbot) {
		Dbuf[pos-bbot] = new&0xff;
		Dbuf[pos-bbot+1] = new>>8;
		return;
	}

	lseek(Tmpdata, (long) pos, SEEK_SET);
	dwrite(Tmpdata, &new, _BPW_);
	lseek(Tmpdata, 0L, SEEK_END);
}


/******************************************************************************
 *	CG UTILITIES
 *****************************************************************************/

void push(i, a) {
	if (Sp >= STKLEN) error("stack overflow");
	Istack[Sp] = i;
	Astack[Sp++] = a;
}


void pop(ip, ap)
int	*ip, *ap;
{
	if (!Sp) error("stack underflow");
	*ip = Istack[--Sp];
	*ap = Astack[Sp];
}


/******************************************************************************
 *	EXTERNAL REFERENCE HANDLING
 *****************************************************************************/


struct extref *findref(id) {
	struct extref	*ep;

	for (ep = Erefs; ep; ep = ep->next) {
		if (ep->id == id && !(ep->name)) return(ep);
	}
	return(NULL);
}


void lx(id) {
	struct extref	*ep;
	int	old;

	if ((ep = findref(id)) == NULL) {
		if ((ep = (struct extref *) malloc(sizeof(struct extref)))
		== NULL)
			nomem();
		ep->id = id;
		ep->offset = BADLBL;
		ep->name = NULL;
		ep->next = Erefs;
		Erefs = ep;
	}
	old = ep->offset;
	ep->offset = Ctop;
	emitw(old);
}


/******************************************************************************
 *	SAVE/LOAD OPERATIONS
 *****************************************************************************/

void stack(how) {
	switch(how) {
		case G_LN:	push(G_LN, num()); break;
		case G_LA:	push(G_LA, num()); break;
		case G_LL:	push(G_LL, laddr()); break;
		case G_LX:	push(G_LX, num()); break;
		case G_LG:	push(G_LG, addr()); break;
	}
}


void load(how, addr) {
	if (Acc_busy) PUSHAX();
	switch(how) {
		case G_LN:	MOVAXN(); emitw(addr); break;
		case G_LA:	MOVAXN(); emitw(arel(addr)); break;
		case G_LL:	MOVAXL(addr); emitv(addr); break;
		case G_LX:	MOVAXG(); lx(addr); break;
		case G_LG:	MOVAXG(); emitw(rel(addr)); break;
		default:	context(0);
	}
	Acc_busy = 1;
}


void loadcx(how, addr) {
	switch(how) {
		case G_LN:	MOVCXN(); emitw(addr); break;
		case G_LA:	MOVCXN(); emitw(arel(addr)); break;
		case G_LL:	MOVCXL(addr); emitv(addr); break;
		case G_LX:	MOVCXG(); lx(addr); break;
		case G_LG:	MOVCXG(); emitw(rel(addr)); break;
		default:	context(1);
	}
}


void flush(n) {
	int	i;

	for (i=0; i<Sp-n; i++) {
		load(Istack[i], Astack[i]);
	}
	if (Sp-n > 0) {
		for (i=0; i<n; i++) {
			Istack[i] = Istack[Sp-n+i];
			Astack[i] = Astack[Sp-n+i];
		}
		Sp = n;
	}
}


void store(op) {
	uint	a;

	flush((op == G_SV)? 1: 0);
	if (op == G_SV) {
		if (Sp) {
			MOVBXAX();
			Acc_busy = 0;
			load(*Istack, *Astack);
			Sp--;
		}
		else {
			POPBX();
		}
		SHLBX1();
	}
	switch(op) {
		case G_SL:	a = laddr(); MOVLAX(a); emitv(a); break;
		case G_SX:	MOVGAX(); lx(num()); break;
		case G_SG:	MOVGAX(); emitw(rel(addr())); break;
		case G_SV:	MOVPAX(); break;
	}
}


/******************************************************************************
 *	BINARY OPERATOR CODE GENERATION
 *****************************************************************************/

void _add(m, a) {
	if (m == G_LN)	{ ADDN(); emitw(a); return; }
	if (m == G_LA)	{ ADDN(); emitw(arel(a)); return; }
	if (m == G_LL)	{ ADDL(a); emitv(a); return; }
	if (m == G_LG)	{ ADDG(); emitw(rel(a)); return; }
	if (m == G_LX)	{ ADDG(); lx(a); return; }
	context(2);
}


void _sub(m, a) {
	if (m == G_LN)	{ SUBN(); emitw(a); return; }
	if (m == G_LA)	{ SUBN(); emitw(arel(a)); return; }
	if (m == G_LL)	{ SUBL(a); emitv(a); return; }
	if (m == G_LG)	{ SUBG(); emitw(rel(a)); return; }
	if (m == G_LX)	{ SUBG(); lx(a); return; }
	context(2);
}


void _mul(m, a) {
	loadcx(m, a);
	MULCX();
}


void _div(m, a) {
	loadcx(m, a);
	XORDXDX();
	DIVCX();
}


void _mod(m, a) {
	loadcx(m, a);
	XORDXDX();
	DIVCX();
	MOVAXDX();
}


void _and(m, a) {
	if (m == G_LN)	{ ANDN(); emitw(a); return; }
	if (m == G_LA)	{ ANDN(); emitw(arel(a)); return; }
	if (m == G_LL)	{ ANDL(a); emitv(a); return; }
	if (m == G_LG)	{ ANDG(); emitw(rel(a)); return; }
	if (m == G_LX)	{ ANDG(); lx(a); return; }
	context(3);
}


void _or(m, a) {
	if (m == G_LN)	{ ORN(); emitw(a); return; }
	if (m == G_LA)	{ ORN(); emitw(arel(a)); return; }
	if (m == G_LL)	{ ORL(a); emitv(a); return; }
	if (m == G_LG)	{ ORG(); emitw(rel(a)); return; }
	if (m == G_LX)	{ ORG(); lx(a); return; }
	context(4);
}


void _xor(m, a) {
	if (m == G_LN)	{ XORN(); emitw(a); return; }
	if (m == G_LA)	{ XORN(); emitw(arel(a)); return; }
	if (m == G_LL)	{ XORL(a); emitv(a); return; }
	if (m == G_LG)	{ XORG(); emitw(rel(a)); return; }
	if (m == G_LX)	{ XORG(); lx(a); return; }
	context(5);
}


void _shl(m, a) {
	loadcx(m, a);
	SHLCL();
}


void _shr(m, a) {
	loadcx(m, a);
	SHRCL();
}


void _rol(m, a) {
	loadcx(m, a);
	ROLCL();
}


void _ror(m, a) {
	loadcx(m, a);
	RORCL();
}


void mkflag(op) {
	switch (op) {
		case G_EQ:	JE(); break;
		case G_NE:	JNE(); break;
		case G_GT:	JG(); break;
		case G_LT:	JL(); break;
		case G_GE:	JGE(); break;
		case G_LE:	JLE(); break;
		case G_UGT:	JA(); break;
		case G_ULT:	JB(); break;
		case G_UGE:	JAE(); break;
		case G_ULE:	JBE(); break;
	}
	emitb(4);
	XORAX();
	JMPS(); emitb(3);
	MOVAXN(); emitw(-1);
}


void comp(op, m, a) {
	if (m == G_LN)		{ CMPN(); emitw(a); }
	else if (m == G_LA)	{ CMPN(); emitw(arel(a)); }
	else if (m == G_LL)	{ CMPL(a); emitv(a); }
	else if (m == G_LG)	{ CMPG(); emitw(rel(a)); }
	else if (m == G_LX)	{ CMPG(); lx(a); }
	else context(6);
	mkflag(op);
}


void binary2(op) {
	Sp = 0;
	POPCX();
	switch(op) {
		case G_ADD:	ADDCX(); break;
		case G_SUB:	XCHG(); SUBCX(); break;
		case G_MUL:	MULCX(); break;
		case G_DIV:	XCHG(); XORDXDX(); DIVCX(); break;
		case G_MOD:	XCHG(); XORDXDX(); DIVCX(); MOVAXDX(); break;
		case G_AND:	ANDCX(); break;
		case G_OR:	ORCX(); break;
		case G_XOR:	XORCX(); break;
		case G_SHL:	XCHG(); SHLCL(); break;
		case G_SHR:	XCHG(); SHRCL(); break;
		case G_ROL:	XCHG(); ROLCL(); break;
		case G_ROR:	XCHG(); RORCL(); break;
		case G_LT: case G_LE: case G_GT: case G_GE:
		case G_UGT: case G_UGE: case G_ULT: case G_ULE:
		case G_EQ: case G_NE:
				CMPCXAX(); mkflag(op); break;
	}
}


void binary(op) {
	int	mode, addr;

	if (Sp < 1) {
		binary2(op); return;
	}

	flush(1);
	Sp--;
	mode = *Istack;
	addr = *Astack;

	switch(op) {
		case G_ADD:	_add(mode, addr); break;
		case G_SUB:	_sub(mode, addr); break;
		case G_MUL:	_mul(mode, addr); break;
		case G_DIV:	_div(mode, addr); break;
		case G_MOD:	_mod(mode, addr); break;
		case G_AND:	_and(mode, addr); break;
		case G_OR:	_or(mode, addr); break;
		case G_XOR:	_xor(mode, addr); break;
		case G_SHL:	_shl(mode, addr); break;
		case G_SHR:	_shr(mode, addr); break;
		case G_ROL:	_rol(mode, addr); break;
		case G_ROR:	_ror(mode, addr); break;
		case G_EQ: case G_NE: case G_LT: case G_LE:
		case G_GT: case G_GE:
		case G_UGT: case G_UGE: case G_ULT: case G_ULE:
				comp(op, mode, addr); break;
	}
}


/******************************************************************************
 *	UNARY OPERATOR CODE GENERATION
 *****************************************************************************/

void unary(op) {
	if (Sp) {
		flush(1);
		Sp--;
		load(*Istack, *Astack);
	}

	switch(op) {
		case G_NEG:	NEGAX(); break;
		case G_COM:	NOTAX(); break;
		case G_LV:	SHLAX1(); MOVBXAX(); MOVAXP(); break;
		case G_ADJ:	MOVBXBP(); SHRBX1(); SUBBXAX(); MOVAXBX();
				break;
		case G_NOT:	NEGAX(); SBBAX(); NOTAX(); break;
	}
}


/******************************************************************************
 *	JUMPS AND BRANCHES
 *****************************************************************************/

struct label *_getlabel(base, id)
struct label	*base;
int	id;
{
	struct label	*lp;

	for (lp = base; lp; lp = lp->next) {
		if (lp->id == id) return(lp);
	}
	return(NULL);
}


void resolve(id, addr, code)
int	id;
uint	addr;
int	code;
{
	struct label	*marks, *mp, **mpp;
	uint	offset, next;

	marks = code? Cmarks: Dmarks;

	if ((mp = _getlabel(marks, id)) == NULL) return;
	offset = mp->offset;
	while (offset != BADLBL) {
		if (code) {
			next = c_extract(offset);
			c_adjust(offset, mkdisp(offset, addr));
		}
		else {
			next = d_extract(offset);
			d_adjust(offset, addr);
		}
		offset = next;
	}

	if (code)
		mpp = &Cmarks;
	else
		mpp = &Dmarks;
	for (mp = marks; mp; ) {
		if (mp->id == id) {
			*mpp = mp->next;
			free(mp);
			return;
		}
 		mpp = &mp->next;
		mp = mp->next;
	}
}


void addlabel() {
	int	id = num();
	struct label	*new;

	if (_getlabel(Labels, id)) error("duplicate label");
	if ((new = (struct label *) malloc(sizeof(struct label))) == NULL)
		nomem();
	new->next = Labels;
	new->id = id;
	new->offset = Ctop;
	Labels = new;
	resolve(new->id, new->offset, 0);
	resolve(new->id, new->offset, 1);
}


uint getlabel(id, code, ok)
int	id, code, *ok;
{
	struct label    *lp;
	uint	old;

	if ((lp = _getlabel(Labels, id)) == NULL) {
		if ((lp = _getlabel(code? Cmarks: Dmarks, id)) == NULL) {
			if ((lp = (struct label *)
			malloc(sizeof(struct label))) == NULL)
				nomem();
			lp->id = id;
			lp->offset = BADLBL;
			if (code) {
				lp->next = Cmarks;
				Cmarks = lp;
			}
			else {
				lp->next = Dmarks;
				Dmarks = lp;
			}
		}
		old = lp->offset;
		lp->offset = code? Ctop: Dtop;
		*ok = 0;
		return(old);
	}
	else {
		*ok = 1;
		return(lp->offset);
	}
}


void jump(op) {
	int	offset, exists, lbl = num();

	if (op == G_JTR) {
		ORAX(); JE(); emitb(3);
	}
	else if (op == G_JFA) {
		ORAX(); JNE(); emitb(3);
	}
	JMP();
	offset = getlabel(lbl, 1, &exists);
	if (exists) offset = mkdisp(Ctop, offset);
	emitw(offset);
	Acc_busy = 0;
}


void clrlabels() {
	struct label	*lp, *old;

	for (lp = Labels; lp; ) {
		old = lp;
		lp = lp->next;
		free(old);
	}
	Labels = NULL;
	if (Cmarks || Dmarks) error("There are undefined labels");
}


static uchar	Switch_code[] = {
			0xEB, 0x14,	/* jmp skip */
	/* again: */	0x8B, 0x14,	/* mov dx,[si] */
			0x46, 0x46,	/* inc si; inc si */
			0x8B, 0x1C,	/* mov bx,[si] */
			0x0B, 0xDB,	/* or  bx,bx */
			0x74, 0x08,	/* jz  default */
			0x46, 0x46,	/* inc si; inc si */
			0x3B, 0xD0,	/* cmp dx, ax */
			0x75, 0xF0,	/* jnz again */
			0xFF, 0xE3,	/* jmp bx */
	/* default: */	0xFF, 0xE2	/* jmp dx */
	/* skip: */
};

void gen_switch() {
	int	i;

	Switch_entry = Ctop+2;
	for (i=0; i<sizeof(Switch_code); i++)
		emitb(Switch_code[i]);
}


void mkswitch() {
	int	i, val, exists, ncase, dflt;
	uint	ptable, bot;

#ifndef FORCE_SWITCH
	if (Switch_entry == BADLBL) gen_switch();
#endif

	flush(0);
	ncase = num();

	bot = addr()+Dframe;
	while (Dtop < bot) demit(0);
	ptable = Dtop;

	dflt = num();

	for (i=0; i<ncase; i++) {
		demit(num());
		val = getlabel(num(), 0, &exists);
	/*	if (exists) val += Cframe;	*/
		demit(val);
	}
	val = getlabel(dflt, 0, &exists);
	/*	if (exists) val += Cframe;	*/
	demit(val);
	demit(0);

	MOVSIN(); emitw(ptable);
	JMP(); emitw(mkdisp(Ctop, Switch_entry));
}


/******************************************************************************
 *	MISC. INSTRUCTIONS
 *****************************************************************************/

void unvec() {
	uint	v;

	if ((v = addr()) == 0) return;
	if (v <= 2*_BPW_) {
		POPCX();
		if (v > _BPW_) POPCX();
	}
	else {
		ADDSPN(v); emitv(v);
	}
}


void vcall() {
	flush(1);
	if (Sp) {
		if (Acc_busy) PUSHAX();
		switch(*Istack) {
			case G_LN:	load(*Istack, *Astack*_BPW_); CALLAX();
					break;
			case G_LA:	load(*Istack, arel(*Astack*_BPW_));
					CALLAX(); break;
			case G_LL:	CALLL(*Astack); emitv(*Astack); break;
			case G_LX:	CALLG(); lx(*Astack); break;
			case G_LG:	CALLG(); emitw(rel(*Astack)); break;
			default:	context(7);
		}
		Sp--;
	}
	else {
		flush(0);
		if (!Acc_busy) POPAX();
		CALLAX();
	}
	Acc_busy = 1;
}


void gen_image() {
	int	n = num();

	while (n--) emitb(num());
}


/******************************************************************************
 *	SYMBOL TABLE CONSTRUCTION
 *****************************************************************************/

char *strsave(s)
char	*s;
{
	char	*new;

	if ((new = (char *) malloc(strlen(s)+1)) == NULL) nomem();
	strcpy(new, s);
	return(new);
}


struct pubsym *findpub(name)
char	*name;
{
	struct pubsym	*pp;

	for (pp = Pubtbl; pp; pp = pp->next) {
		if (!strcmp(pp->name, name)) return(pp);
	}
	return(NULL);
}


struct extref *findext(name)
char	*name;
{
	struct extref	*ep;

	for (ep = Erefs; ep; ep = ep->next) {
		if (ep->name && !strcmp(ep->name, name)) return(ep);
	}
	return(NULL);
}


uint eo_alist(base)
uint	base;
{
	uint	pos, old = base;

	pos = c_extract(base);
	while (pos != BADLBL) {
		old = pos;
		pos = c_extract(pos);
	}
	return(old);
}


void ext_delete(old)
struct extref	*old;
{
	struct extref	*ep, **epp;

	epp = &Erefs;
	ep = Erefs;
	while (ep) {
		if (ep == old) {
			*epp = ep->next;
			if (old->name) free(old->name);
			free(old);
			return;
		}
		epp = &ep->next;
		ep = ep->next;
	}
}


void mkexttbl() {
	int	i, val, n = num();
	uint	alink;
	char	*name;
	struct extref	*ep, *prev;

	for (i=0; i<n; i++) {
		name = string();
		val = num();
		if ((ep = findref(val)) == NULL) continue;	/**/
		if ((prev = findext(name)) != NULL && ep != prev) {
			alink = eo_alist(ep->offset);
			c_adjust(alink, prev->offset);
			ep->name = prev->name;
			prev->name = NULL;
			ext_delete(prev);
		}
		else {
			ep->name = strsave(name);
		}
	}
}


void mkpubtbl() {
	int	i, val, n = num();
	char	*name;
	struct pubsym	*ps;

	for (i=0; i<n; i++) {
		name = string();
		val = addr();
		if (findpub(name)) error("duplicate public symbol");
		if ((ps = (struct pubsym *) malloc(sizeof(struct pubsym)))
		== NULL)
			nomem();
		ps->name = strsave(name);
		ps->offset = val + Dframe;
		ps->next = Pubtbl;
		Pubtbl = ps;
	}
}


void xresolve() {
	struct extref	*ep, *old;
	struct pubsym	*ps;
	uint	pos, next;

	for (ep = Erefs; ep; ) {
		if ((ps = findpub(ep->name)) != NULL) {
			pos = ep->offset;
			while (pos != BADLBL) {
				next = c_extract(pos);
				c_adjust(pos, ps->offset);
				pos = next;
			}
			old = ep;
			ep = ep->next;
			ext_delete(old);
		}
		else
			ep = ep->next;
	}
}


/******************************************************************************
 *	DATA INITIALIZATION
 *****************************************************************************/

void init_data() {
	int	i, bot, count;
	int	tok, val, exists;

	count = num();

	bot = addr()+Dframe;
	while (Dtop < bot) demit(0);

	for (i=0; i<count; i++) {
		if ((tok = next()) == G_REL) {
			val = num()+Dframe/2;
		}
		else if (tok == G_RLA) {
			val = getlabel(num(), 0, &exists);
			/*	if (exists) val += Cframe;	*/
		}
		else {
			val = Value;
		}
		demit(val);
	}
}


/******************************************************************************
 *	MAIN ROUTINES
 *****************************************************************************/

void process(token) {
	int	v;

	switch (token) {
		case G_LN: case G_LL: case G_LG: case G_LX: case G_LA:
			stack(token); break;

		case G_ADD: case G_SUB: case G_MUL: case G_DIV: case G_MOD:
		case G_AND: case G_OR: case G_XOR: case G_SHL: case G_SHR:
		case G_ROL: case G_ROR: case G_EQ: case G_NE: case G_LT:
		case G_LE: case G_GT: case G_GE: case G_UGT: case G_UGE:
		case G_ULT: case G_ULE:
			binary(token); break;

		case G_NEG: case G_COM: case G_NOT: case G_LV: case G_ADJ:
			unary(token); break;

		case G_SL: case G_SG: case G_SX: case G_SV:
			store(token); break;

		case G_SAV:
		case G_PSH:
			flush(0); PUSHAX(); Acc_busy = (token == G_SAV);
			break;

		case G_LBL:
			flush(0); addlabel(); break;

		case G_JMP: case G_JTR: case G_JFA:
			flush(0); jump(token); break;

		case G_SWI:
			mkswitch(); break;

		case G_VEC:
			flush(0); v = addr(); SUBSPN(v); emitv(v); break;

		case G_UNV:
			flush(0);
			/* FALLTHRU */
		case G_RMA:
			unvec(); break;

		case G_VCA:
			vcall(); break;

		case G_FRM:
			PUSHBP(); MOVBPSP(); Acc_busy = 0; break;

		case G_RTN:
			POPBP(); RET(); break;

		case G_HLT:
			flush(0); HALT(); break;

		case G_DRO:
			flush(0); Sp = Acc_busy = 0; break;

		case G_LIN:
			num(); break;	/* LIN is currently discarded */

		case G_INI:
			init_data(); break;

		case G_END:
			xresolve();
			Dframe += addr();
			Cframe = Ctop;
			if (Module_name) free(Module_name);
			Module_name = NULL;
			clrlabels();
			break;

		case G_IMG:
			gen_image(); break;

		case G_EXT:
			mkexttbl(); break;

		case G_PUB:
			mkpubtbl(); break;

		case G_NAM:
			if (Module_name) free(Module_name);
			Module_name = strsave(string());
			Line = 1;
			break;

		default:
			context(8);
	}
}


void collect() {
	int	token;

	while ((token = next()) != ENDFILE) {
		process(token);
	}
}


void copy(in, out) {
	static char	buf[BUFLEN];
	int	n;

	lseek(in, 0L, SEEK_SET);
	while ((n = read(in, buf, BUFLEN)) > 0)
		dwrite(out, buf, n);
}


void generate() {
	uint	csize = Ctop, dsize = Dtop;
	uint	pages, last;
	uint	exehdr[16];

	while (csize++%SEGMENT) emitb(0x00);
	csize--;
#ifdef FORCE_SWITCH
	c_adjust(sizeof(Switch_code)+1, csize>>4);
#else
	c_adjust(1, csize>>4);
#endif
	dwrite(Tmpcode, Cbuf, Cp);
	dwrite(Tmpdata, Dbuf, Dp);
	pages = csize/512 + dsize/512;
	last = (csize%512 + dsize%512 + 32);
	pages += (last/512+1);
	last %= 512;
	if (!last) pages--;

	strcpy((char *) exehdr, "MZ");	/* EXE magic */
	exehdr[1] = last;		/* size of in last page (bytes) */
	exehdr[2] = pages;		/* number of pages (x512 bytes) */
	exehdr[3] = 1;			/* number of reloc entries */
	exehdr[4] = 2;			/* size of header (x16 bytes) */
	exehdr[5] = 0x0000;		/* MINALLOC */
	exehdr[6] = 0xffff;		/* MAXALLOC */
	exehdr[7] = csize/SEGMENT;	/* initial stack segment (SS) */
	exehdr[8] = 0xfffe;		/* initial stack pointer (SP) */
	exehdr[9] = 0;			/* checksum (0=ignore) */
	exehdr[10] = 0;			/* initial entry point (IP) */
	exehdr[11] = 0;			/* initial code segment (CS) */
	exehdr[12] = 0x1C;		/* offset of relocation table */
	exehdr[13] = 0;			/* overlay ID (0=none) */
	/* ----- static header ends here ----- */
#ifdef FORCE_SWITCH
	exehdr[14] = sizeof(Switch_code)+1;
#else
	exehdr[14] = 1;			/* relocation offset */
#endif
	exehdr[15] = 0;			/* relocation segment */

	dwrite(Outfile, exehdr, sizeof(exehdr));
	copy(Tmpcode, Outfile);
	copy(Tmpdata, Outfile);
}


void report() {
	struct extref	*ep;
	int	l = 0, l2;

	if (!Erefs) return;
	fprintf(stderr, "CG: The following externals are unresolved:\n");
	for (ep = Erefs; ep; ep = ep->next) {
		l2 = strlen(ep->name);
		if (l+=l2 > 72) {
			l = l2; fputc('\n', stderr);
		}
		fprintf(stderr, " %s", ep->name);
	}
	fputc('\n', stderr);
}


main() {
	init();
#ifdef FORCE_SWITCH
	gen_switch();
#endif
	collect();
	generate();
	report();
	end();
	return(Erefs != NULL);
}
