//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	THE EXPRESSION PARSER
//
//	Parses an expression, and builds an extended parse tree.
//
//	Public functions:
//		MKTABLE()	Parse/create a TABLE
//		MKSTRING()	Parse/create a STRING
//		EXPR()		Parse an expression and create
//				an extended parse tree
//

#include( blib.hdr)

#include( defs.hdr)
#include( extrn.hdr)
#include( error.hdr)
#include( scan.hdr)
#include( token.hdr)
#include( glue.hdr)
#include( string.hdr)


extern	newnode, newlabel, findsym,
	findsx, freenode, report,
	link, cut,
	undeclared, match, prepend,
	alloc, next, cexpr, fetch,
	xsymname, xlparen, xrparen,
	stcsize, expect;

forward	expr;


//	MKSTRING()
//	Purpose:	Parse and create a STRING literal.
//	Receives:	-
//	Returns:	The address of the first character in the string.
//
public mkstring() {
	var	i, k = 0;
	var	root;

	while (Strval.k) k += 1;
	root = newnode(G_INI);
	root.N_val = k+1;
	root = prepend(root, G_NUM);
	root.N_val = Dtop;
	for (i=0, k) {
		root = prepend(root, G_NUM);
		root.N_val = Strval.i & 0xff;
	}
	root = prepend(root, G_NUM);
	root.N_val = 0;
	alloc(k+1, 1);
	next();
	return root;
}


//	MKTABLE()
//	Purpose:	Parse a TABLE, starting at its first element,
//			and create it.
//	Receives:	-
//	Returns:	The address of the first table element.
//
//	WARNING:	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//			Nesting tables by more than two levels
//			may break the parser (stack overflow).
//	WARNING:	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//
public mktable(ap) {
	var	tbl.TBLLEN, rel.TBLLEN;
	var	sp;
	var	sname.NAMELEN+1;
	var	i, j;
	var	root, t = 0;

	i = 0;
	while (1) {
		if (i >= TBLLEN) report(ETBLL, FATAL);
		if (Token == ADDROF) {
			next();
			bcopy(Strval, sname, NAMELEN);
			sname.NAMELEN = 0;
			xsymname();
			if ((sp = findsym(sname, 0)) == 0)
				if ((sp = findsym(sname, 1)) == 0)
					sp = undeclared(sname);
			sp.S_flags |= REFD;
			if (sp.S_flags & RVAL) report(ELVAL, ERROR);
			if (sp.S_flags & EXT) report(EXADDR, ERROR);
			if (!(sp.S_flags & STAT)) report(EAADDR, ERROR);
			tbl.i = sp.S_addr;
			rel.i = DATA;
		}
		else if (Token == STRING) {
			tbl.i = Dtop;
			t = link(t, mkstring());
			rel.i = DATA;
		}
		else if (Token == LBRACK) {
			next();
			t = link(t, mktable(tbl+i));
			rel.i = DATA;
		}
		else {
			tbl.i = cexpr();
			rel.i = NONE;
		}
		i += 1;
		if (Token != COMMA) break;
		next();
	}
	match(RBRACK, "]", ERROR);

	root = newnode(G_INI);
	root.N_val = i;
	root = prepend(root, G_NUM);
	.ap = root.N_val = Dtop;
	for (j=0, i) {
		if (rel.j != NONE)
			root = prepend(root, (rel.j == CODE)-> G_RLA, G_REL);
		root = prepend(root, G_NUM);
		root.N_val = tbl.j;
	}
	alloc(i, 1);

	if (t) {
		t = cut(t);
		root = link(t, root);
	}
	return root;
}


//	ORDER(EXP, INI)
//	Purpose:	Extract INI nodes from an expression
//	Receives:	EXP	- Pointer to the expression tree
//				to scan
//			INI	- Pointer to the root node of the
//				tree where INI nodes will be
//				prepended
//	Returns:	A parse for the function call.
//
order(exp, inip) {
	if (!exp) return;
	if (exp.N_token == G_LA && exp.N_left) {
		.inip = link(.inip, exp.N_left);
		exp.N_left = 0;
		return;
	}
	order(exp.N_left, inip);
	order(exp.N_right, inip);
}


//	FNCALL(NP)
//	Purpose:	Parse a function call
//	Receives:	NP	- A pointer to the subtree that computes
//				the function address.
//	Returns:	A parse for the function call.
//
fncall(np) {
	var	root, t, ini = 0;
	var	narg = 0;

	next();
	root = t = 0;
	if (Token != RPAREN) while (1) {
		narg += 1;
		if (t) {
			t.N_left = newnode(G_PSH);
			t = t.N_left;
		}
		else {
			root = t = newnode(G_PSH);
		}
		t.N_right = expr(0);
		order(t.N_right, @ini);
		if (Token != COMMA) break;
		next();
	}
	if (ini) ini = cut(ini);
	xrparen();
	fetch(np);
	root = link(ini, root);
	root = link(root, np);
	root.N_token = G_VCA;
	t = newnode(G_RMA);
	t.N_val = narg;
	t.N_left = root;
	return t;
}


//	PRIMARY()
//	Purpose:	Parse a primary factor.
//			PRIMARY() handles numeric literals, SYMBOLS,
//			structures, STRINGS, TABLES, and grouping.
//	Receives:	-
//	Returns:	A parse tree
//
primary() {
	var	np;
	var	sp;
	var	addr;

	switch (Token) {
		case NUMBER:
			np = newnode(G_LN);
			np.N_val = Tvalue;
			np.N_flags = CONST;
			next();
			break;

		case SYMBOL:
			np = newnode(G_SYM);
			if (sp = findsym(Strval, 0)) {
				np.N_flags = 0;
			}
			else if (sp = findsym(Strval, 1)) {
				np.N_flags = GLOBAL;
			}
			else {
				sp = undeclared(Strval);
			}
			sp.S_flags |= REFD;
			if (sp.S_flags & RVAL) {
				np.N_token = G_LN;
				np.N_val = (sp.S_flags & STCT)->
				stcsize(sp), sp.S_addr;
				np.N_flags |= CONST;
			}
			else {
				np.N_val = sp;
				np.N_flags |= DIRECT;
			}
			next();
			break;

		case STRING:
			np = newnode(G_LA);
			np.N_val = Dtop;
			np.N_left = mkstring();
			np.N_flags = CONST;
			break;

		case LPAREN:
			next();
			np = expr(1);
			np.N_flags = CONST;
			xrparen();
			break;

		case LBRACK:
			next();
			np = newnode(G_LA);
			np.N_left = mktable(@addr);
			np.N_val = addr;
			np.N_flags = CONST;
			break;

		default:
			expect("factor", SEVERE);
	}

	return np;
}


//	STRUCT_XS()
//	Purpose:	Parse a binary indirection.
//	Receives:	-
//	Returns:	A parse tree.
//
struct_xs() {
	var	root, t;

	root = primary();
	if (Token == LPAREN) root = fncall(root);
	while (Token == LV) {
		next();
		fetch(root);
		t = newnode(G_ADD);
		t.N_left = root;
		root = t;
		root.N_right = primary();
		fetch(root.N_right);
		t = newnode(G_LV);
		t.N_left = root;
		root = t;
		root.N_flags = 0;
		if (Token == LPAREN) root = fncall(root);
	}
	return root;
}


//	FACTOR()
//	Purpose:	Parse a factor. This includes the handling of
//			the unary operators . - + ! ~ @.
//	Receives:	-
//	Returns:	a parse tree
//
factor() {
	var	root;

	switch (Token) {
		case LV:
			root = newnode(G_LV);
			next();
			root.N_left = factor();
			fetch(root.N_left);
			root.N_flags = 0;
			break;

		case SUB:
		case NOT:
		case LNOT:
			root = newnode((Token == SUB)-> G_NEG,
			(Token == NOT)-> G_COM, G_NOT);
			next();
			root.N_left = factor();
			fetch(root.N_left);
			root.N_flags = CONST;
			break;

		case ADDROF:
			next();
			root = primary();
			if (!(root.N_flags & DIRECT)) {
				report(ELVAL, ERROR);
				break;
			}
			if (root.N_val.S_flags & EXT) {
				report(EXADDR, ERROR);
				break;
			}
			if (root.N_val.S_flags & STAT) {
				root.N_token = G_LA;
				root.N_val = root.N_val.S_addr;
			}
			else {
				var	t;

				root.N_token = G_LN;
				root.N_val = root.N_val.S_addr;
				t = newnode(G_ADJ);
				t.N_left = root;
				root = t;
			}
			break;

		case ADD:
			next(); root = factor(); break;

		default:
			root = struct_xs();
	}

	return root;
}


//	TERM()
//	Purpose:	Parse a term (operators: * / %).
//	Receives:	-
//	Returns:	a parse tree
//
term() {
	var	root, t;

	root = factor();
	while (1) {
		if (Token == MUL || Token == DIV || Token == MOD) {
			t = newnode((Token == MUL)-> G_MUL,
			(Token == DIV)-> G_DIV, G_MOD);
			next();
			t.N_left = root;
			t.N_right = factor();
			root = t;
			fetch(root.N_left);
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	SUM()
//	Purpose:	Parse a sum (operators: + -).
//	Receives:	-
//	Returns:	a parse tree
//
sum() {
	var	root, t;

	root = term();
	while (1) {
		if (Token == ADD || Token == SUB) {
			t = newnode((Token == ADD)-> G_ADD, G_SUB);
			next();
			t.N_left = root;
			t.N_right = term();
			root = t;
			fetch(root.N_left);
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	SH_ROT()
//	Purpose:	Parse a chain of shift/rotate operations
//			(operators: << >> {{ }}).
//	Receives:	-
//	Returns:	a parse tree
//
sh_rot() {
	var	root, t;
	var	op;

	root = sum();
	while (1) {
		if (Token == SHL || Token == SHR || Token == ROL ||
		Token == ROR) {
			switch(Token) {
				case SHL:	op = G_SHL; break;
				case SHR:	op = G_SHR; break;
				case ROL:	op = G_ROL; break;
				case ROR:	op = G_ROR; break;
			}
			t = newnode(op);
			next();
			t.N_left = root;
			t.N_right = sum();
			root = t;
			fetch(root.N_left);
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	COMP()
//	Purpose:	Parse a chain of comparisons (operators: < > <= >=).
//	Receives:	-
//	Returns:	a parse tree
//
comp() {
	var	root, t;
	var	op, uc = 0;

	root = sh_rot();
	while (1) {
		if (Token == KUNSIGN) {
			next(); uc = 1;
		}

		if (Token == GRTR || Token == LESS || Token == GTEQ ||
		Token == LSEQ) {
			switch(Token) {
				case GRTR:	op = uc-> G_UGT, G_GT; break;
				case LESS:	op = uc-> G_ULT, G_LT; break;
				case GTEQ:	op = uc-> G_UGE, G_GE; break;
				case LSEQ:	op = uc-> G_ULE, G_LE; break;
			}
			t = newnode(op);
			next();
			t.N_left = root;
			t.N_right = sh_rot();
			root = t;
			fetch(root.N_left);
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	EQUATION()
//	Purpose:	Parse a chain of comparisons (operators: == !=).
//	Receives:	-
//	Returns:	a parse tree
//
equation() {
	var	root, t;

	root = comp();
	while (1) {
		if (Token == EQU || Token == NEQ) {
			t = newnode((Token == EQU)-> G_EQ, G_NE);
			next();
			t.N_left = root;
			t.N_right = comp();
			root = t;
			fetch(root.N_left);
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	BINAND()
//	Purpose:	Parse a chain of binary AND operations (&).
//	Receives:	-
//	Returns:	a parse tree
//
binand() {
	var	root, t;

	root = equation();
	while (1) {
		if (Token == AND) {
			t = newnode(G_AND);
			next();
			t.N_left = root;
			t.N_right = equation();
			root = t;
			fetch(root.N_left);
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	BINXOR()
//	Purpose:	Parse a chain of binary XOR operations (^).
//	Receives:	-
//	Returns:	a parse tree
//
binxor() {
	var	root, t;

	root = binand();
	while (1) {
		if (Token == XOR) {
			t = newnode(G_XOR);
			next();
			t.N_left = root;
			t.N_right = binand();
			root = t;
			fetch(root.N_left);
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	BINOR()
//	Purpose:	Parse a chain of binary OR operations (|).
//	Receives:	-
//	Returns:	a parse tree
//
binor() {
	var	root, t;

	root = binxor();
	while (1) {
		if (Token == OR) {
			t = newnode(G_OR);
			next();
			t.N_left = root;
			t.N_right = binxor();
			root = t;
			fetch(root.N_left);
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	LOGAND()
//	Purpose:	Parse a chain of logical AND operations (&&).
//			Create code for short-circuit evaluation.
//	Receives:	-
//	Returns:	An extended parse tree.
//
logand() {
	var	root, t;
	var	dest = -1;

	root = binor();
	while (1) {
		if (Token == LAND) {
			fetch(root);
			next();
			if (dest == -1) {
				t = newnode(G_LBL);
				dest = t.N_val = newlabel();
				t.N_right = root;
				root = t;
				root.N_flags = CONST;
			}
			root.N_token = G_JFA;
			t = newnode(G_LBL);
			t.N_val = dest;
			t.N_left = root;
			root = t;
			root.N_right = binor();
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	LOGOR()
//	Purpose:	Parse a chain of logical OR operations (||).
//			Create code for short-circuit evaluation.
//	Receives:	-
//	Returns:	An extended parse tree.
//
logor() {
	var	root, t;
	var	dest = -1;

	root = logand();
	while (1) {
		if (Token == LOR) {
			fetch(root);
			next();
			if (dest == -1) {
				t = newnode(G_LBL);
				dest = t.N_val = newlabel();
				t.N_right = root;
				root = t;
				root.N_flags = CONST;
			}
			root.N_token = G_JTR;
			t = newnode(G_LBL);
			t.N_val = dest;
			t.N_left = root;
			root = t;
			root.N_right = logand();
			fetch(root.N_right);
			root.N_flags = CONST;
		}
		else
			return root;
	}
}


//	COND()
//	Purpose:	Parse a conditional expression
//			( ternary operator -> , ).
//	Receives:	-
//	Returns:	An extended parse tree.
//
cond() {
	var	root, t;
	var	false, end;

	root = logor();
	if (Token == ARROW) {
		fetch(root);
		next();

		t = newnode(G_JFA);
		false = t.N_val = newlabel();
		t.N_left = root;
		root = t;

		t = newnode(G_JMP);
		end = t.N_val = newlabel();
		t.N_left = root;
		root = t;

		root.N_right = cond();
		fetch(root.N_right);
		match(COMMA, "','", ERROR);

		t = newnode(GLUE);
		t.N_left = root;
		root = t;

		root.N_right = newnode(G_LBL);
		t = root.N_right;

		t.N_val = end;
		t.N_left = newnode(G_LBL);
		t.N_left.N_val = false;
		t.N_right = cond();
		fetch(t.N_right);

		root.N_flags = CONST;
	}

	return root;
}


//	BCODE instructions for use with combined assignment operators 
var	ops = [
		G_ROR, G_ROL, G_SHR, G_SHL, G_XOR, G_OR, G_AND,
		G_MOD, G_DIV, G_MUL, G_SUB, G_ADD
	];

//	ASSIGNMENT()
//	Purpose:	Parse an assignment
//			(operators: = += -= *= /= %= &= |= ^= <<= >>= {{= }}=).
//	Receives:	-
//	Returns:	an extended parse tree.
//
assignment() {
	var	root, t;
	var	op = 0;

	root = cond();
	switch (Token) {
		case MOV:
			if (root.N_flags & CONST) report(ELVAL, ERROR);

			next();

			if (root.N_flags & DIRECT) {
				if (root.N_val.S_flags & EXT) {
					root.N_token = G_SX;
					root.N_val=findsx(root.N_val.S_name);
				}
				else {
					root.N_token = (root.N_flags & GLOBAL
					|| root.N_val.S_flags & STAT)->
					G_SG, G_SL;
					root.N_val = root.N_val.S_addr;
				}
				root.N_left = assignment();
				fetch(root.N_left);
			}
			else {
				if (root.N_token == G_LV) {
					t = root;
					root = root.N_left;
					freenode(t);
				}
				t = newnode(G_SV);
				t.N_left = root;
				root = t;
				root.N_right = assignment();
				fetch(root.N_right);
			}

			break;

		case AADD:	op += 1;
		case ASUB:	op += 1;
		case AMUL:	op += 1;
		case ADIV:	op += 1;
		case AMOD:	op += 1;
		case AAND:	op += 1;
		case AOR:	op += 1;
		case AXOR:	op += 1;
		case ASHL:	op += 1;
		case ASHR:	op += 1;
		case AROL:	op += 1;
		case AROR:
			if (root.N_flags & CONST) report(ELVAL, ERROR);

			next();

			if (root.N_flags & DIRECT) {
				var	lvp;
				var	globl, sx;

				lvp = root.N_val;
				globl = (root.N_flags & GLOBAL) != 0;

				if (lvp.S_flags & EXT) {
					root.N_token = G_LX;
					sx = root.N_val = findsx(lvp.S_name);
				}
				else {
					root.N_token = (globl ||
					lvp.S_flags & STAT)-> G_LG, G_LL;
					root.N_val = root.N_val.S_addr;
				}
				t = newnode(ops.op);
				t.N_left = root;
				t.N_right = assignment();
				fetch(t.N_right);
				root = t;

				if (lvp.S_flags & EXT) {
					t = newnode(G_SX);
					t.N_val = sx;
				}
				else {
					t = newnode((globl ||
					lvp.S_flags & STAT)-> G_SG, G_SL);
					t.N_val = lvp.S_addr;
				}
				t.N_left = root;
				root = t;
			}
			else {
				if (root.N_token == G_LV) {
					t = root;
					root = root.N_left;
					freenode(t);
				}
				t = newnode(G_SAV);
				t.N_right = root;
				root = t;
				t = newnode(G_LV);
				t.N_left = root;
				root = t;

				t = newnode(ops.op);
				t.N_left = root;
				root = t;

				root.N_right = assignment();
				fetch(root.N_right);
				t = newnode(G_SV);
				t.N_left = root;
				root = t;
			}

			break;
	}

	return root;
}


//	EXPR(COMMA_OK)
//	Purpose:	Parse an expression, and create an extended parse
//			tree.
//	Receives:	COMMA_OK	- Comma-operator allowed.
//	Returns:	An extended parse tree.
//
public expr(comma_ok) {
	var	root, t;
	var	first = 1;

	root = assignment();
	fetch(root);

	if (comma_ok) while (1) {
		if (Token != COMMA) break;
		next();
		if (first) {
			t = newnode(GLUE);
			t.N_left = root;
			root = t;
		}
		t = assignment();
		fetch(t);
		root.N_token = G_DRO;
		root = link(root, t);
	}

	return root;
}
