//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	THE MAIN MODULE
//
//	Initialize, run, and shutdown the compiler.
//
//	Public functions: none
//

#include( blib.hdr)

#include( defs.hdr)
#include( globl.hdr)
#include( error.hdr)
#include( token.hdr)
#include( glue.hdr)
#include( string.hdr)


extern	sc_init, ut_init, sy_init, em_init, pa_init;
extern	next, do_emit, t_emit, n_emit, emit_sym, dumpsym;
extern	parse, unreferenced;


//	INIT()
//	Purpose:	Initialize globals, allocate pool memory, and
//			symbol table memory.
//	Receives:	NNOD	- number of nodes to allocate
//	Returns:	-
//
init(nnod) {
	sc_init();
	em_init();
	ut_init(nnod);
	pa_init();
	sy_init(nnod);

	O_debug = 0;
	O_opt = 2;
	O_linfo = 0;
}


//	END()
//	Purpose:	Shut down the compiler (flush buffers,
//			dump symbol table [debug], free memory,
//			close files).
//	Receives:	-
//	Returns:	-
//
end() {
	emit_sym(PUB);
	emit_sym(EXT);
	t_emit(G_END);
	n_emit(Dtop);
	do_emit(0);
	if (O_debug) dumpsym();
}


//	SETOPTS(S)
//	Purpose:	Scan an OPTIONS string, and set the requested
//			compiler switches.
//	Receives:	S	- An options string (the argument to
//				  OPTIONS).
//	Returns:	-
//
public setopts(s) {
	var	flag = 1;
	var	optr = 0;

	while (.s) {
		switch (.s) {
			case '+':
				flag = -1; break;
			case '-':
				flag = 0; break;
			case 'd': case 'D':
				O_debug = flag; optr = @O_debug; break;
			case 'l': case 'L':
				O_linfo = flag; optr = @O_linfo; break;
			case 'o': case 'O':
				O_opt = flag; optr = @O_opt; break;

			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				if (optr) .optr = (.s)-'0'; break;
		}
		s += 1;
	}
}


//	START()
//	Purpose:	startup.
//	Receives:	nothing.
//	Returns:	hopefully.
//
start() {
	var	memp.MEMTOP;
	var	strp.STRSPACE;

	Memory = memp;
	Str_pool = strp;

	init(DFLNODES);
	parse();
	end();
	return (Errcount || O_debug)-> 1, 0;
}
