//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	BASIC STRING OPERATIONS
//
//	PUBLIC FUNCTIONS:
//		BCOPY()		copy a memory region
//		STRCPY()	copy a string
//		STREQU()	compare two strings
//		STRCHR()	search for a character in a string
//		STRCAT()	concatenate two strings
//		WRITES()	pack a string and write it to SYSOUT
//		ITOA()		convert a number to ASCII
//

#include( blib.hdr)


//	BCOPY(S, D, L)
//	Purpose:	Copy a memory block
//	Receives:	S	- Source pointer
//			D	- Destination pointer
//			L	- Block length
//	Returns:	-
//
public bcopy(s, d, l) {
	while (l) {
		l -= 1;
		d.l = s.l;
	}
}


//	STRCPY(D, S)
//	Purpose:	Copy a NUL-terminated string
//	Receives:	D	- Destination pointer
//			S	- Source pointer
//	Returns:	-
//
public strcpy(d, s) {
	var	l = 0;

	while (s.l) l += 1;
	bcopy(s, d, l+1);
}


//	STRCAT(S1, S2)
//	Purpose:	Concatenate two NUL-terminated strings (append
//			S1 to S2)
//	Receives:	S1, S2	- strings to concatenate
//	Returns:	-
//
public strcat(s1, s2) {
	while (.s1) s1 += 1;
	strcpy(s1, s2);
}


//	STREQU(S1, S2)
//	Purpose:	Compare two NUL-terminated strings
//	Receives:	S1, S2	- strings to compare
//	Returns:	 0 -> the strings are equal
//			-1 -> the strings are not equal
//
public strequ(s1, s2) {
	while (.s1 || .s2) {
		if (.s1 != .s2) return 0;
		s1 += 1;
		s2 += 1;
	}
	return -1;
}


//	STRCHR(S, C)
//	Purpose:	Search for character C in NUL-string S
//	Receives:	S	- string to scan
//			C	- character to search
//	Returns:	A pointer to the character found, or
//			Null if C is not contained in S
//
public strchr(s, c) {
	while (.s) {
		if (.s == c) return s;
		s += 1;
	}
	return 0;
}


//	WRITES(FD, S)
//	Purpose:	Writes a string to a file
//	Receives:	FD	- file descriptor to write to
//			S	- string to write
//	Returns:	-
//
public writes(fd, s) {
	var	buf.81/_BPW_, l = 0;

	while (s.l) l+=1;
	pack(s, buf);
	writepacked(fd, buf, l);
}


//	ITOA(N, L)
//	Purpose:	Convert a machine word into a string
//	Receives:	N	- number to convert
//			L	- minimum length of string
//				  (will be padded with zeroes)
//	Returns:	A point to a static buffer containing
//			the string.
//
public itoa(n, l) {
	var	i, j, k, t, sign = 0;
	static var	b.12;

	if (n<0) {
		sign = 1;
		n = -n;
	}
	while (n) {
		b.i = n%10+'0';
		n /= 10;
		i += 1;
	}
	if (!i) {
		b.i = '0';
		i += 1;
	}
	if (sign) {
		b.i = '-';
		i += 1;
	}
	while (i<l) {
		b.i = '\s';
		i += 1;
	}
	b.i = 0;
	k = i/2;
	i -= 1;
	for (j=0, k) {
		t = b.j;
		b.j = b.(i-j);
		b.(i-j) = t;
	}
	return(b);
}
