///////////////////////////////////////////////////////////////////////////////
//
//	The BSL Compiler Kit
//	BSL Runtime Library: String Functions
//	Copyright (C) 1995 Nils M. Holm
//
//	VSCAN ( BUFFER, FORMAT, ARGV )
//
//	PURPOSE:	Extract fields from a string.
//	RECEIVES:	BUFFER	- buffer to extract from.
//			FORMAT	- the description of the input fields.
//			ARGV	- a pointer to an argument vector in
//				which the extracted fields will be
//				placed in. The elements of the vector
//				are assumed to be pointer themselves.
//	RETURNS:	The number of fields actually extracted.
//
//	DESCRIPTION:	VSCAN() uses the same field descriptions as
//			VFORMAT() (see there). Each character that
//			is not part of a description will be skipped 
//			on input. In fact, any number of occurences
//			of the given character will be read and
//			discarded.
//			An additional description exists for reading
//			character *classes*:
//
//				%[<range>]
//
//			This descriptor accepts all characters specified
//			in <range>. <range> is a set of characters where
//			sequences like a...z may be written as a-z.
//

#include( blib.hdr)
#include( ../opt.hdr)

extern	bscan, stob, toupper, bcopy;

public vscan(buffer, format, argv) {
	var	flen,
		field.256, class.256, i, j,
		count = 0;

	while (.format) {
		if (.format != '%') {
			while (.buffer == .format) buffer += 1;
			format += 1;
			loop;
		}
		format += 1;
		if (.format == '%') {
			while (.buffer == '%') buffer += 1;
			format += 1;
			loop;
		}
		flen = 0;
		while ('0' <= .format && .format <= '9') {
			flen = flen*10 + .format-'0';
			format += 1;
		}
		if (!flen) flen = 255;
		i = 0;
		if (bscan("BDOXbdox", .format, 8) && .buffer == '-') {
			.field = '-';
			i = 1;
			buffer += 1;
		}
		if (.format == 'b' || .format == 'B') {
			while (.buffer == '0' || .buffer == '1' && i < flen) {
				field.i = .buffer; i += 1;
				buffer += 1;
			}
			field.i = 0;
			..argv = stob(field, 2);
		}
		else if (.format == 'd' || .format == 'D') {
			while ('0' <= .buffer && .buffer <= '9' && i < flen) {
				field.i = .buffer; i += 1;
				buffer += 1;
			}
			field.i = 0;
			..argv = stob(field, 10);
		}
		else if (.format == 'o' || .format == 'O') {
			while ('0' <= .buffer && .buffer <= '7' && i < flen) {
				field.i = .buffer; i += 1;
				buffer += 1;
			}
			field.i = 0;
			..argv = stob(field, 8);
		}
		else if (.format == 'x' || .format == 'X') {
			var	c = toupper(.buffer);
			while (('0' <= c && c <= '9' || 'A' <= c && c <= 'F')
			&& i < flen) {
				field.i = .buffer; i += 1;
				c = toupper(.(buffer += 1));
			}
			field.i = 0;
			..argv = stob(field, 16);
		}
		else if (.format == 'c' || .format == 'C') {
			..argv = .buffer;
			buffer += 1;
		}
		else if (.format == 's' || .format == 'S') {
			while (.buffer && i<flen) {
				field.i = .buffer; i += 1;
				buffer += 1;
			}
			field.i = 0;
			bcopy(field, .argv, i+1);
		}
		else if (.format == '[') {
			j = 0;
			format += 1;
			while (.format && .format != ']') {
				if (.format == '-' && j && .(format+1) &&
				.(format+1) != ']') {
					var	k;

					format += 1;
					for (k=.(format-2)+1, .format+1) {
						class.j = k; j += 1;
					}
				}
				else {
					class.j = .format; j += 1;
				}
				format += 1;
			}
			while (bscan(class, .buffer, j)) {
				field.i = .buffer; i += 1;
				buffer += 1;
			}
			field.i = 0;
			bcopy(field, .argv, i+1);
		}
		argv += 1;
		count += 1;
		format += 1;
	}

	return count;
}
