;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;	The BSL Compiler Kit
;;	BSL/86 Runtime Support: Extended Startup Module
;;	Copyright (C) 1995 Nils M. Holm
;;
;; ----------------------------------------------------------------------------
;;
;;	The extended startup module builds a C-style ARGC/ARGV pair from
;;	the command line options string passed to it.
;;	It expands DOS-style wildcards to file names.
;;	Expansion of wildcard characters may be supressed by using quote
;;	characters (either " or '). The quote characters themselves are
;;	not included in the argument string. To include a quote, it must
;;	be enclosed with quotes of the *other* type:
;;
;;		"this is a quote char --> ' <--"
;;	gives
;;
;;		this is a quote char --> ' <--
;;
;;	Blanks or tabs are used to separate arguments. Both, blanks and
;;	tabs are preserved between quote characters.
;;
;;	The argument text and the vector table together may not occupy
;;	more than ARGVECL - 44 bytes (see below). Only the amount of
;;	memory actually required to build the argument vector is
;;	allocated by the startup module.
;;
;;	***** NOTE *****
;;	The strings contained in the argument vector are PACKED.
;;
;; ----------------------------------------------------------------------------

	.MODEL TINY
	.CODE

ADJUST	equ	16h-100h		;; MAGIC VALUE, DO NOT CHANGE !!!

ARGVECL	equ	4096
ARGVEC	equ	10000h-ARGVECL+44
DTA	equ	10000h-ARGVECL
DIRENT	equ	DTA+001Eh
ATTRIB	equ	DTA+0015h


	ORG	100h

;;
;;	THE INITIAL ENTRY POINT IS HERE
;;
start:	mov	ax,0		;; DATA SEGMENT address will be inserted
	mov	ds,ax		;; at CS:0001, so DON'T CHANGE THIS !!!

	mov	ax,4a00h	;; shrink program space to 2 segments
	mov	bx,8192
	int	21h		;; must succeed because of header info

	jmp	usrcode		;; Add your code at the label 'usrcode'

;;
;;	STRUCTURE REGISTERS {
;;		AX, BX, CX, DX, SI, DI, BP, SP, FLAGS
;;		CS, DS, ES, SS;
;;	};
;;

	EVEN

;;
;;	__SYS is a BSL function. It is assigned the pointer at DS:0
;;	in B0TAIL.BCO.
;;
__SYS:	push	bp			;; create stack frame
	mov	bp,sp

	mov	ax,[bp+4]		;; get interrupt ID
	mov	byte ptr cs:[patch+ADJUST+1],al	;; patch INT instr.

	mov	bp,[bp+6]		;; get pointer to parameter field
	shl	bp,1			;; adjust word address

	mov	ax,[bp]			;; load registers from REGISTERS
	mov	bx,[bp+2]		;; structure (see above)
	mov	cx,[bp+4]
	mov	dx,[bp+6]
	mov	si,[bp+8]
	mov	di,[bp+10]

	push	bp			;; save parameter field
patch:	int	00			;; OPERAND WILL BE PATCHED
	pop	bp

	mov	[bp],ax			;; save registers + flags
	mov	[bp+2],bx
	mov	[bp+4],cx
	mov	[bp+6],dx
	mov	[bp+8],si
	mov	[bp+10],di
	mov	[bp+12],bp
	mov	[bp+14],sp
	lahf
	mov	al,ah
	xor	ah,ah
	mov	[bp+16],ax
	mov	[bp+18],cs
	mov	[bp+20],ds
	mov	[bp+22],es
	mov	[bp+24],ss

	mov	ax,[bp+16]		;; return flags in AX

	pop	bp			;; restore stack frame
	ret


Copyright db	"< BSL86 B0, Copyright (C) 1995 Nils M. Holm >"


;;
;;	LOWCASE -- convert an ASCIZ-string to lower case.
;;
lowcase:
	push	si
	mov	si,ax

dolowc:
	mov	al,[si]
	inc	si
	cmp	al,'A'
	jb	nxtlowc
	cmp	al,'Z'
	ja	nxtlowc
	add	al,32
	mov	[si-1],al
nxtlowc:
	or	al,al
	jnz	dolowc

endlowc:
	pop	si
	ret


;;	GLOB -- expand wildcards
;;
;;	Uses the string at [BX+2] as wildcard, and appends
;;	expanded file names to the argument vector (passed
;;	to it in DI).
;;
glob:	push	cx			;; save quote flag
	push	si			;; save parameter ptr

	mov	si,[bx+2]
	cmp	byte ptr [si],'.'
	jnz	nodot
	pop	si
	pop	cx
	ret

nodot:
	xor	cx,cx			;; match count

	push	cx
	mov	ax,4e00h		;; Find First File
	mov	dx,[bx+2]		;; wildcard == last arg
	mov	cx,003fh		;; all attributes
	int	21h
	pop	cx
	jc	nomatch

trynext:
	test	byte ptr ds:[ATTRIB],0ah
	jnz	skipfile
	mov	al,ds:[DIRENT]
	cmp	al,'.'
	jz	skipfile

nxtmatch:
	or	cx,cx
	jnz	domatch
	inc	bx			;; back up argvec ptr
	inc	bx
	mov	di,[bx]

domatch:
	inc	cx
	test	di,1
	jz	aligned
	inc	di
aligned:
	mov	[bx],di			;; save arg ptr
	dec	bx
	dec	bx

	mov	si,DIRENT		;; copy DIRENT to string pool
cpnxt:	cmp	di,bx
	jae	overflow
	mov	al,[si]
	mov	[di],al
	inc	si
	inc	di
	or	al,al
	jnz	cpnxt

	mov	ax,[bx+2]
	call	lowcase

skipfile:
	mov	ax,4F00h		;; Find Next File
	int	21h
	jnc	trynext

nomatch:
	pop	si			;; restore pmtr ptr
	pop	cx			;; restore quote flag
	ret


;;
;;	OVERFLOW -- Report an argument buffer overflow and abort
;;
msg001	db	"b0: too many arguments", 13, 10, '$'
overflow:
	mov	ax,0900h
	mov	dx,offset msg001 + ADJUST
	push	cs
	pop	ds
	int	21h
	mov	ax,4CFFh
	int	21h

;;
;;	BUILD_ARGV -- Expand wildcards, and build the ARGC/ARGV pair.
;;	Returns ARGV in AX
;;	and	the end of the string pool in DI.
;;
build_argv:
	pop	bp
	sub	sp,ARGVECL		;; alloc argument string pool/vector

	mov	dx,DTA			;; DX points to DOS DTA
	mov	ax,1a00h		;; DOS: set DTA
	int	21h

	mov	di,ARGVEC		;; DI points to string pool
	mov	si,80h			;; SI points to argument string
	mov	bx,0fffeh		;; BX points to end of argvec
	xor	cx,cx			;; quote flag

isspace:
	inc	si			;; fetch char from arg string
nextarg:
	mov	al,es:[si]
	cmp	al,0dh			;; end of arguments?
	jz	endargv

	cmp	al,20h			;; skip over white space
	jz	isspace
	cmp	al,09h
	jz	isspace

	test	di,1
	jz	aligned2
	inc	di
aligned2:
	mov	[bx],di			;; save pointer to next arg
	dec	bx
	dec	bx			;; in argvec

	or	cx,cx			;; already quoting?
	jnz	copyin

	cmp	al,"'"			;; quote character?
	jz	doquote
	cmp	al,'"'
	jnz	copyin
doquote:
	mov	cl,al			;; begin quoted arg
	inc	si			;; skip over quote char
	mov	al,es:[si]
	cmp	al,0dh
	jz	endargv

copyin:
	cmp	di,bx			;; out of space for args
	jae	overflow

	mov	[di],al			;; copy an argument from
	inc	di			;; the parameter string to
	inc	si			;; the string pool
	mov	al,es:[si]
	cmp	al,0dh			;; end of arguments?
	jz	endarg
	or	cx,cx			;; quoting -> ignore white space
	jz	unquoted

	cmp	al,cl
	jz	endarg
	jmp	copyin

unquoted:
	cmp	al,20h			;; white space -> next arg
	jz	endarg
	cmp	al,09h
	jnz	copyin

endarg:
	mov	byte ptr [di],0		;; delimit argument
	inc	di
	or	cx,cx			;; are we quoting?
	jnz	isquoted
	call	glob			;; glob last arg
	jmp	nextarg

isquoted:
	xor	cx,cx			;; end quoted argument
	inc	si			;; skip over quote char
	jmp	nextarg

endargv:
	mov	word ptr [bx],0		;; delimit argvec

	push	bx
	push	di

	mov	di,0fffeh
	mov	dx,di
	sub	dx,bx
	shr	dx,1
	shr	dx,1
	inc	dx

revargs:
	mov	ax,[di]
	mov	cx,[bx]
	mov	[bx],ax
	mov	[di],cx
	dec	di
	dec	di
	inc	bx
	inc	bx
	dec	dx
	jnz	revargs

	pop	di
	pop	ax
	jmp	bp


;;
;;	RELOCATE -- Relocate the argument vector
;;	Receives
;;		AX - pointer to argument vector (end of gap)
;;		DI - end of the string pool (beginning of gap)
;;
relocate:
	pop	bp

	mov	cx,di		;; CX = size of argvec
	sub	cx,ARGVEC

	sub	ax,di		;; DI = destination for arg pool
	mov	di,ax
	add	di,ARGVEC
	add	di,cx
	dec	di

	mov	si,ARGVEC	;; SI = arg pool
	add	si,cx
	dec	si

	inc	cx
	test	cx,1
	jnz	doreloc
	dec	di

doreloc:
	std
	rep	movsb

	pop	si
	and	di,0fffeh
	mov	sp,di
	push	si
	mov	dx,ax
nxtreloc:
	mov	ax,[si]
	or	ax,ax
	jz	endreloc
	add	ax,dx
	shr	ax,1
	mov	[si],ax
	inc	si
	inc	si
	jmp	nxtreloc
endreloc:
	jmp	bp


;;
;;	Used-defined startup function may be called here
;;
usrcode:
	call	build_argv

	push	ds
	pop	es

	push	ax
	call	relocate

	pop	ax
	shr	ax,1
	push	ax

	shl	ax,1
	neg	ax
	shr	ax,1
	dec	ax
	push	ax

	end	start

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
